/* DVB USB framework compliant Linux driver for the
*	TeVii S630
*
* Copyright (C) 2008 Bob Liu (Bob@Turbosight.com)
* Igor M. Liplianin (liplianin@me.by)
*
*	This program is free software; you can redistribute it and/or modify it
*	under the terms of the GNU General Public License as published by the
*	Free Software Foundation, version 2.
*
* see Documentation/dvb/README.dvb-usb for more information
*/
#include <linux/version.h>
#include "tevii-usb.h"
#include "mt312.h"



#ifndef USB_PID_TEVIIUSB
#define USB_PID_TEVIIUSB 0xd630
#endif


#define TEVIIUSB_READ_MSG 0
#define TEVIIUSB_WRITE_MSG 1

/* on my own*/
#define TEVIIUSB_VOLTAGE_CTRL (0x1800)
#define TEVIIUSB_RC_QUERY (0x1a00)

struct teviiusb_state {
	u32 last_key_pressed;
};
struct teviiusb_rc_keys {
	u32 keycode;
	u32 event;
};

/* debug */
static int dvb_usb_teviiusb_debug;
module_param_named(debug, dvb_usb_teviiusb_debug, int, 0644);
MODULE_PARM_DESC(debug, "set debugging level (1=info 2=xfer (or-able))." DVB_USB_DEBUG_STATUS);

DVB_DEFINE_MOD_OPT_ADAPTER_NR(adapter_nr);

static int teviiusb_op_rw(struct usb_device *dev, u8 request, u16 value,
			u16 index, u8 * data, u16 len, int flags)
{
	int ret;
	u8 u8buf[len];

	unsigned int pipe = (flags == TEVIIUSB_READ_MSG) ?
				usb_rcvctrlpipe(dev, 0) : usb_sndctrlpipe(dev, 0);
	u8 request_type = (flags == TEVIIUSB_READ_MSG) ? USB_DIR_IN : USB_DIR_OUT;

	if (flags == TEVIIUSB_WRITE_MSG)
		memcpy(u8buf, data, len);
	ret = usb_control_msg(dev, pipe, request, request_type | USB_TYPE_VENDOR,
				value, index , u8buf, len, 2000);

	if (flags == TEVIIUSB_READ_MSG)
		memcpy(data, u8buf, len);
	return ret;
}

/* I2C */
static int teviiusb_i2c_transfer(struct i2c_adapter *adap, struct i2c_msg msg[],
		int num)
{
struct dvb_usb_device *d = i2c_get_adapdata(adap);
	int i = 0, ret = 0;
	u8 buf6[20];
	u8 inbuf[20];
	

	if (!d)
		return -ENODEV;
	if (mutex_lock_interruptible(&d->i2c_mutex) < 0)
		return -EAGAIN;

	switch (num) {
	case 2:
		
		buf6[0]=msg[1].len;//lenth
		buf6[1]=msg[0].addr;//demod addr
		buf6[2]=msg[0].buf[0];//register
		
		ret = teviiusb_op_rw(d->udev, 0x90, 0, 0,
					buf6, 3, TEVIIUSB_WRITE_MSG);
		msleep(5);
		ret = teviiusb_op_rw(d->udev, 0x91, 0, 0,
					inbuf, 1, TEVIIUSB_READ_MSG);
		memcpy(msg[1].buf, inbuf, msg[1].len);
		break;
	case 1:
		switch (msg[0].addr) {
		case 0xd0:
		case 0x1c:
			/* write to stv0299 register */
			buf6[0] = msg[0].len+1;//lenth
			buf6[1] = msg[0].addr;//demod addr
			for(i=0;i<msg[0].len;i++)
				{
				buf6[2+i] = msg[0].buf[i];//register
				}
			ret = teviiusb_op_rw(d->udev, 0x80, 0, 0,
					buf6, msg[0].len+2, TEVIIUSB_WRITE_MSG);
			msleep(3);
			break;
		case 0xc2:
		case 0xc0:
			if (msg[0].flags == 0) {
			/* write to tuner pll */
				buf6[0] = msg[0].len+1;//lenth
				buf6[1] = msg[0].addr;//tuner addr
				for(i=0;i<msg[0].len;i++)
				{
				buf6[2+i] = msg[0].buf[i];//register
				}
				ret = teviiusb_op_rw(d->udev, 0x80, 0, 0,
						buf6, msg[0].len+2, TEVIIUSB_WRITE_MSG);
			} 
			msleep(3);
			break;
		case (TEVIIUSB_RC_QUERY):
			ret  = teviiusb_op_rw(d->udev, 0xb8, 0, 0,
					buf6, 4, TEVIIUSB_READ_MSG);
			msg[0].buf[0] = buf6[1];
			msg[0].buf[1] = buf6[2];
			msleep(3);
			break;
		case (TEVIIUSB_VOLTAGE_CTRL):
			buf6[0] = 3;
			buf6[1] = msg[0].buf[0];
			ret = teviiusb_op_rw(d->udev, 0x8a, 0, 0,
					buf6, 2, TEVIIUSB_WRITE_MSG);
			msleep(3);
			break;
		}

		break;
	}

	mutex_unlock(&d->i2c_mutex);
	return num;
}

static u32 teviiusb_i2c_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C;
}

static struct mt312_config zl313_config = {
	.demod_address = 0x1c,
};


static struct i2c_algorithm teviiusb_i2c_algo = {
	.master_xfer = teviiusb_i2c_transfer,
	.functionality = teviiusb_i2c_func,
};


static int zl10039_tuner_set_params(struct dvb_frontend* fe, struct dvb_frontend_parameters *params)
{
	u8 buf[4];
	u16 div;
	struct i2c_msg msg = { .addr = 0xc0, .flags = 0, .buf = buf, .len = sizeof(buf) };
	struct dvb_usb_adapter *udev_adap =
		(struct dvb_usb_adapter *)(fe->dvb->priv);
	
	printk("zl10039_tuner_set_params %d\n",params->frequency);
	//div = (params->frequency + (125/2)) / 125;
	div = (u16)((params->frequency*10) / 10111 );


	if (fe->ops.i2c_gate_ctrl)
		fe->ops.i2c_gate_ctrl(fe, 1);

	buf[0] = 0xf;//reg index
	buf[1] = 0x40;
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);
	
	buf[0] = 0xf;//reg index
	buf[1] = 0;
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);
	
	buf[0] = 0x2;//reg index
	buf[1] = (0x02 << 4) | 0x0A;
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);

	buf[0] = 0x6;//reg index
	buf[1] = 0x0a;
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);

	buf[0] = 0x4;//reg index
	buf[1] = 0xee;
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);

	buf[0] = 0x7;//reg index
	buf[1] = (0x06 << 3);
	msg.len = 2;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);
	
	buf[0] = 0x0;//reg index
	buf[1] = 0x05;
	buf[2] = 0xfd;
	msg.len = 3;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);

	buf[0] = 0x0;//reg index
	buf[1] = (u8)(div >> 8);
	buf[2] = (u8)(div & 0x00FF);
	msg.len = 3;
	if (i2c_transfer(&udev_adap->dev->i2c_adap, &msg, 1) != 1)
		return -EIO;
	msleep(10);
	

	
	return 0;
}




static int teviiusb_read_mac_address(struct dvb_usb_device *d, u8 mac[6])
{
	int i,ret;
	u8 ibuf[3] = {0, 0,0};
	u8 eeprom[256], eepromline[16];

	for (i = 0; i < 256; i++) {
		ibuf[0]=1;//lenth
		ibuf[1]=0xa0;//eeprom addr
		ibuf[2]=i;//register
		ret = teviiusb_op_rw(d->udev, 0x90, 0, 0,
					ibuf, 3, TEVIIUSB_WRITE_MSG);
		ret = teviiusb_op_rw(d->udev, 0x91, 0, 0,
					ibuf, 1, TEVIIUSB_READ_MSG);
			if (ret < 0) {
				err("read eeprom failed.");
				return -1;
			} else {
				eepromline[i%16] = ibuf[0];
				eeprom[i] = ibuf[0];
			}
			
			if ((i % 16) == 15) {
				deb_xfer("%02x: ", i - 15);
				debug_dump(eepromline, 16, deb_xfer);
			}
	}
	memcpy(mac, eeprom + 16, 6);
	return 0;
};

static int teviiusb_set_voltage(struct dvb_frontend *fe, fe_sec_voltage_t voltage)
{
	static u8 command_13v[1] = {0x00};
	static u8 command_18v[1] = {0x01};
	struct i2c_msg msg[] = {
		{.addr = TEVIIUSB_VOLTAGE_CTRL, .flags = 0,
			.buf = command_13v, .len = 1},
	};
	
	struct dvb_usb_adapter *udev_adap =
		(struct dvb_usb_adapter *)(fe->dvb->priv);
	if (voltage == SEC_VOLTAGE_18)
		msg[0].buf = command_18v;
	info("@@teviiusb_set_voltage %d\n",voltage);
	i2c_transfer(&udev_adap->dev->i2c_adap, msg, 1);
	return 0;
}





static struct dvb_usb_device_properties teviiusb_properties;

static int teviiusb_frontend_attach(struct dvb_usb_adapter *d)
{

		d->fe = dvb_attach(mt312_attach, &zl313_config,
					&d->dev->i2c_adap);
		if (d->fe != NULL) {
			d->fe->ops.set_voltage = teviiusb_set_voltage;
			d->fe->ops.tuner_ops.set_params =zl10039_tuner_set_params;
			info("Attached zl10313!\n");
			return 0;
		}
	return -EIO;
}



static struct dvb_usb_rc_key teviiusb_rc_keys[] = {
	{ 0xf8,	0x0a, KEY_Q },		/*power*/
	{ 0xf8,	0x0c, KEY_M },		/*mute*/
	{ 0xf8,	0x11, KEY_1 },
	{ 0xf8,	0x12, KEY_2 },
	{ 0xf8,	0x13, KEY_3 },
	{ 0xf8,	0x14, KEY_4 },
	{ 0xf8,	0x15, KEY_5 },
	{ 0xf8,	0x16, KEY_6 },
	{ 0xf8,	0x17, KEY_7 },
	{ 0xf8,	0x18, KEY_8 },
	{ 0xf8,	0x19, KEY_9 },
	{ 0xf8, 0x10, KEY_0 },
	{ 0xf8, 0x1c, KEY_PAGEUP },	/*ch+*/
	{ 0xf8, 0x0f, KEY_PAGEDOWN },	/*ch-*/
	{ 0xf8, 0x1a, KEY_O },		/*vol+*/
	{ 0xf8, 0x0e, KEY_Z },		/*vol-*/
	{ 0xf8, 0x04, KEY_R },		/*rec*/
	{ 0xf8, 0x09, KEY_D },		/*fav*/
	{ 0xf8, 0x08, KEY_BACKSPACE },	/*rewind*/
	{ 0xf8, 0x07, KEY_A },		/*fast*/
	{ 0xf8, 0x0b, KEY_P },		/*pause*/
	{ 0xf8, 0x02, KEY_ESC },	/*cancel*/
	{ 0xf8, 0x03, KEY_G },		/*tab*/
	{ 0xf8, 0x00, KEY_UP },		/*up*/
	{ 0xf8, 0x1f, KEY_ENTER },	/*ok*/
	{ 0xf8, 0x01, KEY_DOWN },	/*down*/
	{ 0xf8, 0x05, KEY_C },		/*cap*/
	{ 0xf8, 0x06, KEY_S },		/*stop*/
	{ 0xf8, 0x40, KEY_F },		/*full*/
	{ 0xf8, 0x1e, KEY_W },		/*tvmode*/
	{ 0xf8, 0x1b, KEY_B },		/*recall*/

};



static int teviiusb_rc_query(struct dvb_usb_device *d, u32 *event, int *state)
{
	struct teviiusb_state *st = d->priv;
	u8 key[2];
	struct i2c_msg msg[] = {
		{.addr = TEVIIUSB_RC_QUERY, .flags = I2C_M_RD, .buf = key,
		.len = 2},
	};
	int i;

	*state = REMOTE_NO_KEY_PRESSED;
	if (teviiusb_i2c_transfer(&d->i2c_adap, msg, 1) == 1) {
		for (i = 0; i < ARRAY_SIZE(teviiusb_rc_keys); i++) {
			if (teviiusb_rc_keys[i].data == msg[0].buf[0]) {
				*state = REMOTE_KEY_PRESSED;
				*event = teviiusb_rc_keys[i].event;
				st->last_key_pressed =
					teviiusb_rc_keys[i].event;
				break;
			}
		st->last_key_pressed = 0;
		}
	}
	/* info("key: %x %x\n",key[0],key[1]); */
	return 0;
}

static struct usb_device_id teviiusb_table[] = {
	{USB_DEVICE(0x9022, 0xd630)},
	{ }
};

MODULE_DEVICE_TABLE(usb, teviiusb_table);

static int teviiusb_load_firmware(struct usb_device *dev,
			const struct firmware *frmwr)
{
	u8 *b, *p;
	int ret = 0, i;
	u8 reset;
	const struct firmware *fw;
	const char *filename = "dvb-usb-teviiusb-s630.fw";
	switch (dev->descriptor.idProduct) {
	case 0xd630:
		ret = request_firmware(&fw, filename, &dev->dev);
		if (ret != 0) {
			err("did not find the firmware file. (%s) "
			"Please see linux/Documentation/dvb/ for more details "
			"on firmware-problems.", filename);
			return ret;
		}
		break;
	default:
		fw = frmwr;
		break;
	}
	info("start downloading TEVIIUSB firmware");
	p = kmalloc(fw->size, GFP_KERNEL);
	reset = 1;
	/*stop the CPU*/
	teviiusb_op_rw(dev, 0xa0, 0x7f92, 0, &reset, 1, TEVIIUSB_WRITE_MSG);
	teviiusb_op_rw(dev, 0xa0, 0xe600, 0, &reset, 1, TEVIIUSB_WRITE_MSG);

	if (p != NULL) {
		memcpy(p, fw->data, fw->size);
		for (i = 0; i < fw->size; i += 0x40) {
			b = (u8 *) p + i;
			if (teviiusb_op_rw(dev, 0xa0, i, 0, b , 0x40,
					TEVIIUSB_WRITE_MSG) != 0x40) {
				err("error while transferring firmware");
				ret = -EINVAL;
				break;
			}
		}
		/* restart the CPU */
		reset = 0;
		if (ret || teviiusb_op_rw(dev, 0xa0, 0x7f92, 0, &reset, 1,
					TEVIIUSB_WRITE_MSG) != 1) {
			err("could not restart the USB controller CPU.");
			ret = -EINVAL;
		}
		if (ret || teviiusb_op_rw(dev, 0xa0, 0xe600, 0, &reset, 1,
					TEVIIUSB_WRITE_MSG) != 1) {
			err("could not restart the USB controller CPU.");
			ret = -EINVAL;
		}

		msleep(100);
		kfree(p);
	}
	return ret;
}

static struct dvb_usb_device_properties teviiusb_properties = {
	.caps = DVB_USB_IS_AN_I2C_ADAPTER,
	.usb_ctrl = DEVICE_SPECIFIC,
	.firmware = "dvb-usb-teviiusb-s630.fw",
	.size_of_priv = sizeof(struct teviiusb_state),
	.no_reconnect = 1,

	.i2c_algo = &teviiusb_i2c_algo,
	.rc_key_map = teviiusb_rc_keys,
	.rc_key_map_size = ARRAY_SIZE(teviiusb_rc_keys),
	.rc_interval = 150,
	.rc_query = teviiusb_rc_query,

	.generic_bulk_ctrl_endpoint = 0x81,
	/* parameter for the MPEG2-data transfer */
	.num_adapters = 1,
	.download_firmware = teviiusb_load_firmware,
	.read_mac_address = teviiusb_read_mac_address,
		.adapter = {
		{
			.frontend_attach = teviiusb_frontend_attach,
			.streaming_ctrl = NULL,
			//.tuner_attach = teviiusb_earda_tuner_attach,
			.stream = {
				.type = USB_BULK,
				.count = 8,
				.endpoint = 0x82,
				.u = {
					.bulk = {
						.buffersize = 4096,
					}
				}
			},
		}
	},
	.num_device_descs = 1,
	.devices = {
		{"TeVii DVBS USB2.0",
			{&teviiusb_table[0], NULL},
			{NULL},
		}
	}
};

static int teviiusb_probe(struct usb_interface *intf,
		const struct usb_device_id *id)
{
	if (0 == dvb_usb_device_init(intf, &teviiusb_properties,
			THIS_MODULE, NULL, adapter_nr)) {
		return 0;
	}
	return -ENODEV;
}

static struct usb_driver teviiusb_driver = {
	.name = "teviiusb",
	.probe = teviiusb_probe,
	.disconnect = dvb_usb_device_exit,
	.id_table = teviiusb_table,
};

static int __init teviiusb_module_init(void)
{
	int ret =  usb_register(&teviiusb_driver);
	if (ret)
		err("usb_register failed. Error number %d", ret);

	return ret;
}

static void __exit teviiusb_module_exit(void)
{
	usb_deregister(&teviiusb_driver);
}

module_init(teviiusb_module_init);
module_exit(teviiusb_module_exit);

MODULE_AUTHOR("Bob Liu <Bob@turbosight.com>");
MODULE_DESCRIPTION("Driver for TeVii S630 USB DVBS");
MODULE_VERSION("0.1");
MODULE_LICENSE("GPL");
