/* DVB USB framework compliant Linux driver for the
*	Tevii S660
*
* Copyright (C) 2008 Bob Liu (Bob@Turbosight.com)
* Igor M. Liplianin (liplianin@me.by)
*
*	This program is free software; you can redistribute it and/or modify it
*	under the terms of the GNU General Public License as published by the
*	Free Software Foundation, version 2.
*
* see Documentation/dvb/README.dvb-usb for more information
*/

/* 
* History:
*
* March 2009 Konstantin Dimitrov <kosio.dimitrov@gmail.com>
* 
* 
*/

#include <linux/version.h>
#include "teviis660.h"
#include "ds3000.h"


#ifndef USB_PID_TEVIIS660_1
#define USB_PID_TEVIIS660_1 0xD660
#endif

#define TEVIIS660_READ_MSG 0
#define TEVIIS660_WRITE_MSG 1

/* on my own*/
#define TEVIIS660_VOLTAGE_CTRL (0x1800)
#define TEVIIS660_RC_QUERY (0x1a00)

struct teviis660_state {
	u32 last_key_pressed;
};
struct teviis660_rc_keys {
	u32 keycode;
	u32 event;
};

/* debug */
static int dvb_usb_teviis660_debug;
module_param_named(debug, dvb_usb_teviis660_debug, int, 0644);
MODULE_PARM_DESC(debug, "set debugging level (1=info 2=xfer (or-able))." DVB_USB_DEBUG_STATUS);

DVB_DEFINE_MOD_OPT_ADAPTER_NR(adapter_nr);

static int teviis660_op_rw(struct usb_device *dev, u8 request, u16 value,
			u16 index, u8 * data, u16 len, int flags)
{
	int ret;
	u8 u8buf[len];

	unsigned int pipe = (flags == TEVIIS660_READ_MSG) ?
				usb_rcvctrlpipe(dev, 0) : usb_sndctrlpipe(dev, 0);
	u8 request_type = (flags == TEVIIS660_READ_MSG) ? USB_DIR_IN : USB_DIR_OUT;

	if (flags == TEVIIS660_WRITE_MSG)
		memcpy(u8buf, data, len);
	ret = usb_control_msg(dev, pipe, request, request_type | USB_TYPE_VENDOR,
				value, index, u8buf, len, 2000);

	if (flags == TEVIIS660_READ_MSG)
		memcpy(data, u8buf, len);
	return ret;
}

/* I2C */
static int teviis660_i2c_transfer(struct i2c_adapter *adap, struct i2c_msg msg[],
		int num)
{
struct dvb_usb_device *d = i2c_get_adapdata(adap);
	int i = 0, ret = 0;
	u8 ibuf[1], obuf[3];
	u8 buf6[20];

	if (!d)
		return -ENODEV;
	if (mutex_lock_interruptible(&d->i2c_mutex) < 0)
		return -EAGAIN;

	switch (num) {
	case 2: {
		/* read */
		obuf[0] = msg[0].len;
		obuf[1] = msg[0].addr<<1;
		obuf[2] = msg[0].buf[0];

		ret = teviis660_op_rw(d->udev, 0x90, 0, 0,
					obuf, 3, TEVIIS660_WRITE_MSG);
		msleep(5);
		ret = teviis660_op_rw(d->udev, 0x91, 0, 0,
					ibuf, 1, TEVIIS660_READ_MSG);
		memcpy(msg[1].buf, ibuf, msg[1].len);
		break;
	}
	case 1:
		switch (msg[0].addr) {
		case 0x68: {
			if (msg[0].buf[0] == 0xb0) {
				/* firmware */
				buf6[0] = 4;
				buf6[1] = msg[0].addr<<1;
				buf6[2] = msg[0].buf[0];
				buf6[3] = msg[0].buf[1];
				buf6[4] = msg[0].buf[2];
				ret = teviis660_op_rw(d->udev, 0x80, 0, 0,
                                                        buf6, 5, TEVIIS660_WRITE_MSG);
                                //msleep(3);
			} else {
				/* write to register */
				buf6[0] = msg[0].len+1;//lenth
				buf6[1] = msg[0].addr<<1;//demod addr
				for(i=0;i<msg[0].len;i++) {
				buf6[2+i] = msg[0].buf[i];//register
				}
				ret = teviis660_op_rw(d->udev, 0x80, 0, 0,
							buf6, msg[0].len+2, TEVIIS660_WRITE_MSG);
				msleep(3);
			}
			break;
		}
		case 0x60: {
			/* write to register */
			buf6[0] = msg[0].len+1;//lenth
			buf6[1] = msg[0].addr<<1;//demod addr
			for(i=0;i<msg[0].len;i++) {
				buf6[2+i] = msg[0].buf[i];//register
			}
			ret = teviis660_op_rw(d->udev, 0x80, 0, 0,
						buf6, msg[0].len+2, TEVIIS660_WRITE_MSG);
			msleep(3);

			break;
		}
		case (TEVIIS660_RC_QUERY): {
			ret  = teviis660_op_rw(d->udev, 0xb8, 0, 0,
					buf6, 4, TEVIIS660_READ_MSG);
			msg[0].buf[0] = buf6[2];
			msg[0].buf[1] = buf6[3];
			msleep(3);
			//info("RC_QUERY %x %x %x %x\n",buf6[0],buf6[1],buf6[2],buf6[3]);
			break;
		}
		case (TEVIIS660_VOLTAGE_CTRL): {
			buf6[0] = 3;
			buf6[1] = msg[0].buf[0];
			ret = teviis660_op_rw(d->udev, 0x8a, 0, 0,
					buf6, 2, TEVIIS660_WRITE_MSG);
			
			break;
		}
		}

		break;
	}

	mutex_unlock(&d->i2c_mutex);
	return num;
}

static u32 teviis660_i2c_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C;
}



static struct i2c_algorithm teviis660_i2c_algo = {
	.master_xfer = teviis660_i2c_transfer,
	.functionality = teviis660_i2c_func,
};

static struct ds3000_config teviis660_ds3000_config = {
		.demod_address = 0x68,
};

static int teviis660_read_mac_address(struct dvb_usb_device *d, u8 mac[6])
{
	int i,ret;
	u8 ibuf[3] = {0, 0,0};
	u8 eeprom[256], eepromline[16];

	for (i = 0; i < 256; i++) {
		ibuf[0]=1;//lenth
		ibuf[1]=0xa0;//eeprom addr
		ibuf[2]=i;//register
		ret = teviis660_op_rw(d->udev, 0x90, 0, 0,
					ibuf, 3, TEVIIS660_WRITE_MSG);
		ret = teviis660_op_rw(d->udev, 0x91, 0, 0,
					ibuf, 1, TEVIIS660_READ_MSG);
			if (ret < 0) {
				err("read eeprom failed");
				return -1;
			} else {
				eepromline[i%16] = ibuf[0];
				eeprom[i] = ibuf[0];
			}
			
			if ((i % 16) == 15) {
				deb_xfer("%02x: ", i - 15);
				debug_dump(eepromline, 16, deb_xfer);
			}
	}
	memcpy(mac, eeprom + 16, 6);
	return 0;
};

static int teviis660_set_voltage(struct dvb_frontend *fe, fe_sec_voltage_t voltage)
{
	static u8 command_13v[1] = {0x00};
	static u8 command_18v[1] = {0x01};
	struct i2c_msg msg[] = {
		{.addr = TEVIIS660_VOLTAGE_CTRL, .flags = 0,
			.buf = command_13v, .len = 1},
	};
	
	struct dvb_usb_adapter *udev_adap =
		(struct dvb_usb_adapter *)(fe->dvb->priv);
	if (voltage == SEC_VOLTAGE_18)
		msg[0].buf = command_18v;
	//info("teviis660_set_voltage %d",voltage);
	i2c_transfer(&udev_adap->dev->i2c_adap, msg, 1);
	return 0;
}

static struct dvb_usb_device_properties teviis660_properties;

static int teviis660_frontend_attach(struct dvb_usb_adapter *d)
{
	if ((d->fe = dvb_attach(ds3000_attach, &teviis660_ds3000_config,
					&d->dev->i2c_adap)) != NULL) {
			d->fe->ops.set_voltage = teviis660_set_voltage;
			printk("S660: DS3000 attached.\n");
			return 0;
	}

	return -EIO;
}



static struct dvb_usb_rc_key teviis660_rc_keys[] = {
	{ 0xff,	0x84, KEY_POWER },		/*power*/
	{ 0xff,	0x94, KEY_MUTE },		/*mute*/
	{ 0xff,	0x87, KEY_1 },
	{ 0xff,	0x86, KEY_2 },
	{ 0xff,	0x85, KEY_3 },
	{ 0xff,	0x8b, KEY_4 },
	{ 0xff,	0x8a, KEY_5 },
	{ 0xff,	0x89, KEY_6 },
	{ 0xff,	0x8f, KEY_7 },
	{ 0xff,	0x8e, KEY_8 },
	{ 0xff,	0x8d, KEY_9 },
	{ 0xff, 0x92, KEY_0 },
	{ 0xff, 0x86, KEY_CHANNELUP },	/*ch+*/
	{ 0xff, 0x91, KEY_CHANNELDOWN },	/*ch-*/
	{ 0xff, 0x93, KEY_VOLUMEUP },		/*vol+*/
	{ 0xff, 0x8c, KEY_VOLUMEDOWN },		/*vol-*/
	{ 0xff, 0x83, KEY_RECORD },		/*rec*/
	{ 0xff, 0x98, KEY_PAUSE  },		/*pause*/
	{ 0xff, 0x99, KEY_OK },	/*ok*/
	{ 0xff, 0x9a, KEY_SHUFFLE },		  /* snapshot */
	{ 0xff, 0x81, KEY_UP },
	{ 0xff, 0x90, KEY_LEFT },
	{ 0xff, 0x82, KEY_RIGHT },
	{ 0xff, 0x88, KEY_DOWN },
	{ 0xff, 0x95, KEY_FAVORITES },
	{ 0xff, 0x97, KEY_SUBTITLE },
	{ 0xff, 0x9d, KEY_ZOOM },
	{ 0xff, 0x9f, KEY_EXIT },
	{ 0xff, 0x9e, KEY_MENU },
	{ 0xff, 0x9c, KEY_EPG },
	{ 0xff, 0x80, KEY_PREVIOUS },
	{ 0xff, 0x9d, KEY_MODE }


};



static int teviis660_rc_query(struct dvb_usb_device *d, u32 *event, int *state)
{
	struct teviis660_state *st = d->priv;
	u8 key[2];
	struct i2c_msg msg[] = {
		{.addr = TEVIIS660_RC_QUERY, .flags = I2C_M_RD, .buf = key,
		.len = 2},
	};
	int i;

	*state = REMOTE_NO_KEY_PRESSED;
	if (teviis660_i2c_transfer(&d->i2c_adap, msg, 1) == 1) {
		//info("key: %x %x\n",msg[0].buf[0],msg[0].buf[1]); 
		for (i = 0; i < ARRAY_SIZE(teviis660_rc_keys); i++) {
			if (teviis660_rc_keys[i].data == msg[0].buf[0]) {
				*state = REMOTE_KEY_PRESSED;
				*event = teviis660_rc_keys[i].event;
				st->last_key_pressed =
					teviis660_rc_keys[i].event;
				break;
			}
		st->last_key_pressed = 0;
		}
	}
	 
	return 0;
}

static struct usb_device_id teviis660_table[] = {
	{USB_DEVICE(0x9022, 0xd660)},
	{USB_DEVICE(USB_VID_CYPRESS, USB_PID_TEVIIS660_1)},
	{ }
};

MODULE_DEVICE_TABLE(usb, teviis660_table);

static int teviis660_load_firmware(struct usb_device *dev,
			const struct firmware *frmwr)
{
	u8 *b, *p;
	int ret = 0, i;
	u8 reset;
	const struct firmware *fw;
	const char *filename = "dvb-usb-teviis660.fw";
	switch (dev->descriptor.idProduct) {
	case 0xd660:
		ret = request_firmware(&fw, filename, &dev->dev);
		if (ret != 0) {
			err("did not find the firmware file. (%s) "
			"Please see linux/Documentation/dvb/ for more details "
			"on firmware-problems.", filename);
			return ret;
		}
		break;
	default:
		fw = frmwr;
		break;
	}
	info("start downloading TEVIIS660 firmware");
	p = kmalloc(fw->size, GFP_KERNEL);
	reset = 1;
	/*stop the CPU*/
	teviis660_op_rw(dev, 0xa0, 0x7f92, 0, &reset, 1, TEVIIS660_WRITE_MSG);
	teviis660_op_rw(dev, 0xa0, 0xe600, 0, &reset, 1, TEVIIS660_WRITE_MSG);

	if (p != NULL) {
		memcpy(p, fw->data, fw->size);
		for (i = 0; i < fw->size; i += 0x40) {
			b = (u8 *) p + i;
			if (teviis660_op_rw(dev, 0xa0, i, 0, b , 0x40,
					TEVIIS660_WRITE_MSG) != 0x40) {
				err("error while transferring firmware");
				ret = -EINVAL;
				break;
			}
		}
		/* restart the CPU */
		reset = 0;
		if (ret || teviis660_op_rw(dev, 0xa0, 0x7f92, 0, &reset, 1,
					TEVIIS660_WRITE_MSG) != 1) {
			err("could not restart the USB controller CPU.");
			ret = -EINVAL;
		}
		if (ret || teviis660_op_rw(dev, 0xa0, 0xe600, 0, &reset, 1,
					TEVIIS660_WRITE_MSG) != 1) {
			err("could not restart the USB controller CPU.");
			ret = -EINVAL;
		}

		msleep(100);
		kfree(p);
	}
	return ret;
}

static struct dvb_usb_device_properties teviis660_properties = {
	.caps = DVB_USB_IS_AN_I2C_ADAPTER,
	.usb_ctrl = DEVICE_SPECIFIC,
	.firmware = "dvb-usb-teviis660.fw",
	.size_of_priv = sizeof(struct teviis660_state),
	.no_reconnect = 1,

	.i2c_algo = &teviis660_i2c_algo,
	.rc_key_map = teviis660_rc_keys,
	.rc_key_map_size = ARRAY_SIZE(teviis660_rc_keys),
	.rc_interval = 150,
	.rc_query = teviis660_rc_query,

	.generic_bulk_ctrl_endpoint = 0x81,
	/* parameter for the MPEG2-data transfer */
	.num_adapters = 1,
	.download_firmware = teviis660_load_firmware,
	.read_mac_address = teviis660_read_mac_address,
		.adapter = {
		{
			.frontend_attach = teviis660_frontend_attach,
			.streaming_ctrl = NULL,
			.stream = {
				.type = USB_BULK,
				.count = 8,
				.endpoint = 0x82,
				.u = {
					.bulk = {
						.buffersize = 4096,
					}
				}
			},
		}
	},
	.num_device_descs = 1,
	.devices = {
		{"TEVII S660 DVBS2 USB2.0",
			{&teviis660_table[0], NULL},
			{NULL},
		}
	}
};

static int teviis660_probe(struct usb_interface *intf,
		const struct usb_device_id *id)
{
	if (0 == dvb_usb_device_init(intf, &teviis660_properties,
			THIS_MODULE, NULL, adapter_nr)) {
		return 0;
	}
	return -ENODEV;
}

static struct usb_driver teviis660_driver = {
	.name = "teviis660",
	.probe = teviis660_probe,
	.disconnect = dvb_usb_device_exit,
	.id_table = teviis660_table,
};

static int __init teviis660_module_init(void)
{
	int ret =  usb_register(&teviis660_driver);
	if (ret)
		err("usb_register failed. Error number %d", ret);

	return ret;
}

static void __exit teviis660_module_exit(void)
{
	usb_deregister(&teviis660_driver);
}

module_init(teviis660_module_init);
module_exit(teviis660_module_exit);

MODULE_AUTHOR("Bob Liu <Bob@turbosight.com>");
MODULE_DESCRIPTION("Driver for TEVII S660");
MODULE_VERSION("0.2");
MODULE_LICENSE("GPL");
